"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __spreadArray = (this && this.__spreadArray) || function (to, from) {
    for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
        to[j] = from[i];
    return to;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RETRY_POLLING_MS = void 0;
var react_1 = require("react");
var MqlContext_1 = __importDefault(require("../context/MqlContext/MqlContext"));
var FetchMqlQueryTimeSeries_1 = __importDefault(require("../queries/mql/FetchMqlQueryTimeSeries"));
var MqlQueryTypes_1 = require("../queries/mql/MqlQueryTypes");
var actions_1 = require("./actions");
var useCreateMqlQuery_1 = __importDefault(require("./utils/useCreateMqlQuery"));
var getErrorMessage_1 = __importDefault(require("./utils/getErrorMessage"));
// This is the delay between the _response_ from the last query and the _start_ of the new query
var QUERY_POLLING_MS = 400;
// Length of time to wait before retrying when query fails
exports.RETRY_POLLING_MS = 200;
// Time in milliseconds to wait for a query to respond successfully before showing a slow loading message to the user.
var LONG_FETCH_QUERY_ATTEMPT_MAX = 20000; // 20 seconds
// As we keep track of prevoius queries, we don't want the list to grow unbounded in long running browser sessions
var CANCELLED_QUERY_LIST_LENGTH = 10;
var initialState = {
    retries: 0,
    queryId: null,
    queryStatus: MqlQueryTypes_1.MqlQueryStatus.Pending,
    data: null,
    cancelledQueries: [],
    fetchStartTime: null,
    isTakingForever: false,
    doRetryAfterExpiredQuery: false,
};
/*
    At first glance, it is counter-intuitive that we should have a mix of the Reducer pattern
    and the Hook pattern. The explanation for this complexity is that we have several fetches
    that require coordination.
      1. We post a mutation to create a query
      2. We use the query ID to poll for results
      3. None of this can take place until an MQL Server URL is supplied
    This reducer abstracts all this complexity.
*/
function mqlQueryReducer(state, action) {
    var _a, _b, _c;
    switch (action.type) {
        case "postQueryStart": {
            // This condition is triggered when there is already a queryId being fetched.
            // In this case we want to add "cancel" that query by adding it to the cancelledQueries list
            if (state.queryId) {
                return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Running, cancelledQueries: __spreadArray([
                        state.queryId
                    ], state.cancelledQueries.slice(CANCELLED_QUERY_LIST_LENGTH - 1)), data: null, fetchStartTime: Date.now(), isTakingForever: false, errorMessage: undefined });
            }
            return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Running, fetchStartTime: Date.now(), isTakingForever: false, data: null, errorMessage: undefined, retries: 0 });
        }
        case "postQueryFail": {
            return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Failed, errorMessage: action.errorMessage, isTakingForever: false, fetchStartTime: null });
        }
        case "postQuerySuccess": {
            return __assign(__assign({}, state), { queryId: action.queryId, errorMessage: undefined });
        }
        case "postQueryCachedResultsSuccess": {
            return __assign(__assign({}, state), { queryId: null, queryStatus: MqlQueryTypes_1.MqlQueryStatus.Successful, data: (_b = (_a = action.data) === null || _a === void 0 ? void 0 : _a.createMqlQuery) === null || _b === void 0 ? void 0 : _b.query, fetchStartTime: null, isTakingForever: false, errorMessage: undefined });
        }
        case "fetchResultsRunning": {
            var now = Date.now();
            var diff = now - (state.fetchStartTime || 0);
            if (state.queryStatus !== MqlQueryTypes_1.MqlQueryStatus.Running && state.isTakingForever !== diff >= LONG_FETCH_QUERY_ATTEMPT_MAX && state.errorMessage !== undefined) {
                return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Running, isTakingForever: diff >= LONG_FETCH_QUERY_ATTEMPT_MAX, errorMessage: undefined });
            }
            else {
                return state;
            }
        }
        case "retryFetchResults": {
            var now = Date.now();
            var diff = now - (state.fetchStartTime || 0);
            return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Running, isTakingForever: diff >= LONG_FETCH_QUERY_ATTEMPT_MAX, errorMessage: undefined, retries: state.retries + 1 });
        }
        case "fetchResultsFail": {
            return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Failed, isTakingForever: false, fetchStartTime: null, errorMessage: action.errorMessage, doRetryAfterExpiredQuery: false, retries: 0 });
        }
        case "fetchResultsSuccess": {
            return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Successful, data: (_c = action.data) === null || _c === void 0 ? void 0 : _c.mqlQuery, fetchStartTime: null, isTakingForever: false, errorMessage: undefined });
        }
        case "fetchResultsExpiredQuery": {
            return __assign(__assign({}, state), { queryId: null, queryStatus: initialState.queryStatus, data: null, errorMessage: undefined, doRetryAfterExpiredQuery: true });
        }
        default: {
            throw new Error();
        }
    }
}
// This custom hook consists of two useHooks that should asynchronously handle all scenarios for this chained
// data fetching we are doing. It should do it in a high performance way and in a way that is resilient to race conditions if the
// user updates their request while a query is in flight
function useMqlQuery(_a) {
    var queryInput = _a.queryInput, metricName = _a.metricName, skip = _a.skip, _b = _a.retries, retries = _b === void 0 ? 5 : _b;
    var _c = react_1.useContext(MqlContext_1.default), useQuery = _c.useQuery, handleCombinedError = _c.handleCombinedError, mqlServerUrl = _c.mqlServerUrl;
    var _d = react_1.useReducer(mqlQueryReducer, initialState), state = _d[0], dispatch = _d[1];
    var createMqlQuery = useCreateMqlQuery_1.default({ metricName: metricName, formState: queryInput, dispatch: dispatch, retries: retries }).createMqlQuery;
    react_1.useEffect(function () {
        if (!metricName || !mqlServerUrl || skip) {
            return;
        }
        dispatch({ type: "postQueryStart" });
        createMqlQuery({ stateRetries: state.retries });
    }, [queryInput, metricName, mqlServerUrl, skip]);
    var _skip = skip ||
        !state.queryId ||
        state.cancelledQueries.includes(state.queryId); /* && !isRunning*/
    var _e = useQuery({
        query: FetchMqlQueryTimeSeries_1.default,
        variables: {
            queryId: state.queryId || ""
        },
        pause: _skip,
    }), _f = _e[0], data = _f.data, error = _f.error, refetchMqlQuery = _e[1];
    var retry = function () {
        setTimeout(function () {
            dispatch({ type: "retryFetchResults" });
            refetchMqlQuery();
        }, exports.RETRY_POLLING_MS);
    };
    react_1.useEffect(function () {
        var _a, _b, _c, _d, _e, _f;
        if (error) {
            // if error is an expired query id, the mql server may have been restarted, restart the entire query.
            if (actions_1.shouldRetryAfterExpiredQuery({
                errorMessage: error === null || error === void 0 ? void 0 : error.message,
                doRetryAfterExpiredQuery: state.doRetryAfterExpiredQuery
            })) {
                dispatch(actions_1.doRetryAfterExpiredQueryAction(retries));
                createMqlQuery({ stateRetries: state.retries });
            }
            else if (retries > 0 && state.retries !== retries) {
                retry();
            }
            else {
                dispatch({
                    type: "fetchResultsFail",
                    errorMessage: getErrorMessage_1.default(error),
                });
                handleCombinedError(error);
            }
            return;
        }
        if (!(data === null || data === void 0 ? void 0 : data.mqlQuery)) {
            return;
        }
        var _g = data.mqlQuery, status = _g.status, id = _g.id;
        if (id && state.cancelledQueries.includes(id) /*&& !isRunning*/) {
            return;
        }
        if (status === MqlQueryTypes_1.MqlQueryStatus.Successful) {
            dispatch({
                type: "fetchResultsSuccess",
                data: data,
                handleCombinedError: handleCombinedError,
            });
        }
        if (
        // isRunning ||
        status === MqlQueryTypes_1.MqlQueryStatus.Running ||
            status === MqlQueryTypes_1.MqlQueryStatus.Pending) {
            window.setTimeout(function () {
                dispatch({ type: "fetchResultsRunning" });
                refetchMqlQuery();
            }, QUERY_POLLING_MS);
        }
        if (status === MqlQueryTypes_1.MqlQueryStatus.Failed) {
            if (retries > 0 && state.retries !== retries) {
                retry();
            }
            else {
                if ((_a = data === null || data === void 0 ? void 0 : data.mqlQuery) === null || _a === void 0 ? void 0 : _a.error) {
                    var error_1 = data.mqlQuery.error;
                    dispatch({
                        type: "fetchResultsFail",
                        errorMessage: error_1,
                    });
                    handleCombinedError({
                        name: "Unknown Error",
                        message: error_1,
                        graphQLErrors: [],
                    }, {
                        queryId: (_b = data === null || data === void 0 ? void 0 : data.mqlQuery) === null || _b === void 0 ? void 0 : _b.id,
                        queryStatus: (_c = data === null || data === void 0 ? void 0 : data.mqlQuery) === null || _c === void 0 ? void 0 : _c.status,
                        json: error_1
                    });
                }
                else {
                    var jsonString = void 0;
                    try {
                        jsonString = JSON.stringify((_d = data === null || data === void 0 ? void 0 : data.mqlQuery) === null || _d === void 0 ? void 0 : _d.result);
                    }
                    catch (e) {
                        jsonString = "Invalid data.mqlQuery.result";
                    }
                    var errorMessage = "This query failed for an unknown reason.";
                    dispatch({
                        type: "fetchResultsFail",
                        errorMessage: errorMessage,
                    });
                    handleCombinedError({
                        name: "Unknown Error",
                        message: errorMessage,
                        graphQLErrors: [],
                    }, {
                        queryId: (_e = data === null || data === void 0 ? void 0 : data.mqlQuery) === null || _e === void 0 ? void 0 : _e.id,
                        queryStatus: (_f = data === null || data === void 0 ? void 0 : data.mqlQuery) === null || _f === void 0 ? void 0 : _f.status,
                        json: jsonString
                    });
                }
            }
        }
    }, [data, error, state.cancelledQueries, handleCombinedError]);
    return state;
}
exports.default = useMqlQuery;
